# Copyright (c) 2025, Tom Ouellette
# Licensed under the GNU GPLv3 License

from anci import main

from .brightfield import brightfield

__all__ = ["brightfield"]

if __name__ == "__main__":
    main("bamboo: flexible gigapixel image analysis for the spatial biology era")
